header {
package frontend.parser;
import frontend.symbol.*;
import java.util.LinkedList;
}
class PL0Parser extends Parser;

options {
	defaultErrorHandler=false;  // Don't generate standard error handlers
	buildAST = true;            // Build an Abstract Syntax Tree
	exportVocab=PL0;            // Call Vocabulary PL0
}

tokens {
	EXPR;
	BLOCK;
	CONSTASSIGN;
	VAR;
	CONSTD;
	PROC;
	PROGRAM<AST=frontend.parsetree.ProcedureNode>;

	ASSIGNOP<AST=frontend.parsetree.AssignNode>;

	NUMBER<AST=frontend.parsetree.NumberNode>;
	IDENT<AST=frontend.parsetree.IdentNode>;

	PLUS<AST=frontend.parsetree.BinopNode>;
	MINUS<AST=frontend.parsetree.BinopNode>;
	TIMES<AST=frontend.parsetree.BinopNode>;
	OVER<AST=frontend.parsetree.BinopNode>;

	EQUALS<AST=frontend.parsetree.BinrelNode>;
	DIAMOND<AST=frontend.parsetree.BinrelNode>;
	LT<AST=frontend.parsetree.BinrelNode>;
	LEQ<AST=frontend.parsetree.BinrelNode>;
	GT<AST=frontend.parsetree.BinrelNode>;
	GEQ<AST=frontend.parsetree.BinrelNode>;

	QUESTION<AST=frontend.parsetree.InputNode>;
	EXCLAM<AST=frontend.parsetree.OutputNode>;
}

{
	SymbolTable symbolTable = SymbolTable.getSymbolTable();
	private boolean pendingMinus = false;
	private LinkedList exceptions = new LinkedList();


    //-------------------------------------------------------------------------
    // Symboltable adapter methods
    // The following methods are provided to give a single set of entry
    // calls into the symbol table.  This makes it easy to add debugging
    // code that will track all calls to symbolTable.popScope, for instance
    // These are direct pass-through calls to the symbolTable, but a
    // few have special function.
    //-------------------------------------------------------------------------

    public void popScope() {
    	symbolTable.popScope();
    }

	public LinkedList getExceptions() {
		return exceptions;
	}

    public void defineConst(Token constant, Token value) {
    	Symbol s = new Symbol(constant, Symbol.SYMBOL_CONST);
    	try {
	        symbolTable.defineConst(s, Integer.parseInt(value.getText()));
	    }
	    catch(DoubleDeclarationException e) {
	    	exceptions.add(e.getMessage());
	    }
    }

    public void defineVar(Token var) {
    	Symbol s = new Symbol(var, Symbol.SYMBOL_VAR);
    	try {
	        symbolTable.defineVar(s);
	    }
	    catch(DoubleDeclarationException e) {
	    	exceptions.add(e.getMessage());
	    }
    }

    public void defineProcedure(Token procName) {
    	Symbol s = new Symbol(procName, Symbol.SYMBOL_PROC);
    	try {
	        symbolTable.defineProcedure(s);
	    }
	    catch(DoubleDeclarationException e) {
	    	exceptions.add(e.getMessage());
	    }
    }

    public Definition lookup(Token t) {
    	return symbolTable.lookup(t.getText());
    }

    public void checkForDeclaration(Token t) {
    	Symbol s = new Symbol(t);
    	symbolTable.identifierUsage(s);

    }

    public void checkForVariable(Token t) {
    	Symbol s = new Symbol(t);
    	try {
	    	symbolTable.checkForVariable(s);
	    }
	    catch (NoVariableException nve) {
	    	exceptions.add(nve.getMessage());
	    }
	    catch (NoDeclarationException nde) {
	    	exceptions.add(nde.getMessage());
	    }
    }

    public void checkForProcedure(Token t) {
    	Symbol s = new Symbol(t);
    	try {
	    	symbolTable.checkForProcedure(s);
	    }
	    catch (NoProcedureException e) {
	    	exceptions.add(e.getMessage());
	    }
    }

    public void checkForValue(Token t) {
    	Symbol s = new Symbol(t);
    	try {
	    	symbolTable.checkForValue(s);
	    }
	    catch (NoValueException nve) {
	    	exceptions.add(nve.getMessage());
	    }
	    catch (NoDeclarationException nde) {
	    	exceptions.add(nde.getMessage());
	    }
    }

    public void resolveUnknown() {
    	try {
	    	symbolTable.resolveUnknown();
	    }
	    catch (NoDeclarationException e) {
	    	exceptions.add(e.getMessage());
	    }
    }

	public void checkRange(Token t) throws NumberOutOfRangeException {
		String s = t.getText();
		if (pendingMinus) {
			s = '-' + s;
		}
		try {
			Integer.parseInt(s);
		}
		catch (NumberFormatException e) {
			exceptions.add(new NumberOutOfRangeException(s, t.getLine(), t.getColumn()).getMessage());
		}

		pendingMinus = false;
	}
}



program:		block DOT! {#program = #([PROGRAM, "PROGRAM"], program); resolveUnknown();};

block:			(constdecl!)? (vardecl!)? (procdecl)* statement;


constdecl:		"CONST" constassign (COMMA constassign)* SEMICOLON;
constassign: 	i:IDENT EQUALS n:NUMBER { checkRange(n); defineConst(i, n); };  //factored out for purpose of clearness


vardecl:		"VAR" i1:IDENT {defineVar(i1);} (COMMA i2:IDENT {defineVar(i2);})* SEMICOLON;


procdecl:		"PROCEDURE"^<AST=frontend.parsetree.ProcedureNode> i:IDENT {defineProcedure(i);}
				SEMICOLON! block
				SEMICOLON! {popScope();};



statement:		(i:IDENT {checkForVariable(i);} ASSIGNOP^ expression
				| "CALL"^<AST=frontend.parsetree.CallNode> i2:IDENT {checkForProcedure(i2);}
				| QUESTION^ i3:IDENT {checkForVariable(i3);}
				| EXCLAM^ i4:IDENT {checkForValue(i4);}
				| "BEGIN"^<AST=frontend.parsetree.BeginNode> statement (SEMICOLON! statement)* "END"!
				| "IF"^<AST=frontend.parsetree.IfNode> condition "THEN"! statement
				| "WHILE"^<AST=frontend.parsetree.WhileNode> condition "DO"! statement)?;


condition:		"ODD"^<AST=frontend.parsetree.OddNode> expression
				| expression (EQUALS^ | DIAMOND^ | LT^ | LEQ^ | GT^ | GEQ^) expression;

expression:		(PLUS^<AST=frontend.parsetree.PlusNode> | MINUS^<AST=frontend.parsetree.MinusNode> {pendingMinus = true;})?
				term {pendingMinus = false;}((PLUS^ | MINUS^) term)*;


term:			factor ((TIMES^ | OVER^) factor)*;


factor:			i:IDENT {checkForValue(i);}
				| n:NUMBER {checkRange(n);}
				| LPAREN! expression RPAREN!;





class PL0Lexer extends Lexer;
options {
	defaultErrorHandler=false;
    charVocabulary = '\0'..'\377';	//unicode
	k = 2; 							//2 symbols lookahead because with k=1 there is an ambiguity < / <= and > / >=
}


WS: 				( ' '
				    | '\t'
				    | '\f'
				    // handle newlines
					| 	( "\r\n"  // DOS
						| '\r'    // Mac
				    	| '\n'    // Unix
				    	)  { newline(); }			//if a new line is found, increase internal line number (for error messages)
    				) { $setType(Token.SKIP); };   	//all tokens matched as whitespace should be skipped




DOT:				'.';
COMMA:				',';
SEMICOLON:			';';
ASSIGNOP:			":=";
QUESTION:			'?';
EXCLAM:				'!';
LT:					'<';
LEQ:				"<=";
GT:					'>';
GEQ:				">=";
DIAMOND:			'#';
EQUALS:				'=';
PLUS:				'+';
MINUS:				'-';
TIMES:				'*';
OVER:				'/';
LPAREN:				'(';
RPAREN:				')';



NUMBER:				('0'..'9')+;
IDENT:				('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'_'|'0'..'9')*;