import compiler.*;

import frontend.*;
import backend.*;
import intermediate.*;
import frontend.symbol.*;
import java.util.*;
import java.io.FileNotFoundException;

/**
 * Das Hauptprogramm. Es wird eine PL0-Datei als Eingabe genommen und interpretiert.
 */
public class PL0Interpreter implements compiler.Compiler {

    private final static String errorNoFile = "Es muss eine zu interpretierende Datei angegeben werden.";
    private final static String errorFileNotFound = "Die Datei wurde nicht gefunden: ";
    private final static String errorUsage = "Rufen Sie den Interpreter so auf: java -jar PL0Interpreter.jar <pl0datei>";
    private PL0Frontend frontend;
    private InterpreterBackend backend;

    /**
     * Konstruktor
     */
    public PL0Interpreter() {
        frontend = new PL0Frontend(this);
        backend = new InterpreterBackend(this);
    }

    public Frontend getFrontend() {
        return frontend;
    }

    public Backend getBackend() {
        return backend;
    }


    /**
     * Erzeugen des FrameContainers und Fllen desselben mit linearisiertem Zwischencode
     */
    public FrameContainer setupFrameContainer() {
        FrameContainer fc = new FrameContainer();
        SymbolTable table = SymbolTable.getSymbolTable();
        processScope(table.getBaseScope(), fc, 0);
        return fc;
    }

    /**
     * Rekursives Einfgen des Hauptprogramms und der Prozeduren in den FrameContainer
     */
    private void processScope(ScopeDef sd, FrameContainer fc, int level) {
        SymbolTable table = SymbolTable.getSymbolTable();
        table.pushScope(sd);
        int nrOfDefs = 0;
        int nrOfConsts = 0;
        LinkedList constDefs = new LinkedList();
        ArrayList defs = new ArrayList();
        for (Iterator it = sd.getElements().keySet().iterator(); it.hasNext();) {
            String name = (String)it.next();
            Definition d = table.lookup(name);
            if (d instanceof VariableDef) {
                nrOfDefs++;
            }
            if (d instanceof ConstantDef) {
                nrOfDefs++;
                nrOfConsts++;
                constDefs.add(d);
                defs.add(null);
            }
        }
         //In der obigen Iteration ist nicht sichergestellt, dass die Konstanten in der richtigen
         //Reihenfolge in der Liste landen (Reihenfolge ihrer Definition). Das holen wir
         //hier nach.
        for (Iterator it2 = constDefs.iterator(); it2.hasNext();) {
            ConstantDef cd = (ConstantDef)it2.next();
            int position = cd.getPosInScope();
            defs.set(position, new Integer(cd.getValue()));

        }
        fc.addFrame(new InterpreterFrame(nrOfDefs, nrOfConsts, defs, level), sd.getScopeNumber());

        for (Iterator itScope = sd.scopeIdentifiers(); itScope.hasNext();) {
            //ScopeDef sd = (ScopeDef)itScope.next();
            processScope((ScopeDef)itScope.next(), fc, level + 1);
        }
        table.popScope();

    }


    public static void main(String[] args) {
        PL0Interpreter interpreter = new PL0Interpreter();
        if (args.length == 0) {
            System.err.println(errorNoFile);
            System.err.println(errorUsage);
            System.exit(-1);
        }
        interpreter.getFrontend().setInputSource(args[0]);
        CodeContainer cc = null;
        try {
            cc = interpreter.getFrontend().transformToIntermediate();
        }
        catch (Exception e) {
            if (e.getClass() == FileNotFoundException.class) {
                System.err.println(errorFileNotFound + args[0]);
                System.exit(-2);
            }
        }
        if (!interpreter.getFrontend().exceptionsFound()) {
            FrameContainer fc = interpreter.setupFrameContainer();
            LinearImm lin = new LinearImm();
            List codelist = lin.createLinearImmList(cc);
            ImmContainer ic = new ImmContainer(codelist);
            interpreter.getBackend().transformToExecutable(codelist, fc);
        }
    }


}
