package frontend.symbol;

import java.util.*;

/**
 * Es gibt Definitionen, die einen neuen Sichtbarkeitsbereich erffnen (Prozeduren). Sie mssen Buch fhren
 * knnen, welche Definitionen sich in ihnen befinden. Diese Funktionalitt wird von ScopeDef als
 * abstrakter Oberklasse bereitgestellt.
 */
public abstract class ScopeDef extends Definition {

    private int scopeNumber;

    /**
     * In diesem Sichtbarkeitsbereich enthaltene Definitionen.
     */
    private Hashtable elements;

    private int nrOfDef = 0;

    private LinkedList unknown = new LinkedList();

    /**
     * Konsturktor
     * @param name Name der Definition
     * @param occ Vorkommen der Definition im Quelltext
     * @param parentScope Sichbarkeitsbereich, zu dem diese Definition gehrt.
     */
    public ScopeDef(String name, Occurrence occ, ScopeDef parentScope) {
        super(name, occ, parentScope);
        elements = new Hashtable();
        scopeNumber = SymbolTable.getScopeCounter();
    }

    protected void addUnknown(Definition def) {
        unknown.add(def);
    }


    /**
     * Fgt eine Definition diesem Sichtbarkeitsbereich hinzu. Falls es schon eine gibt, liegt eine
     * doppelte Definition vor und es wird eine DoubleDeclarationException geworfen.
     */
    protected void add(Definition def){
        //schau nach, ob diese Definition in diesem Sichbarkeitsbereich schon existiert.
        Definition oldDef = (Definition)elements.get(def.getName());
        if (oldDef != null) {        // Wenn ja => Doppeldefinition
            throw new DoubleDeclarationException(def, oldDef);
        }
        elements.put(def.getName(), def); //=> ansonsten hinzufgen
        def.setParentScope(this);
        //Speichere in jeder Konstanten- und Variablendefinition, welche Position sie innerhalb der
        //Symboltabelle hat, diese Information wird bentigt, wenn man bei der Zwischencodeerzeugung
        //die Offsets fr Variablen und Konstanten bentigt. Die Hashmap kann die Reihenfolge, in der
        //die Definitionen hinzugefgt wurden, nicht garantieren (This class makes no guarantees as to
        //the order of the map; in particular, it does not guarantee that the order will remain constant
        //over time.)
        if (def instanceof ConstantDef || def instanceof VariableDef) {
            def.setPosInScope(nrOfDef);
            nrOfDef++;
        }
    }


    /**
     * Gibt an, ob es in diesem Sichtbarkeitsberich schon Definitionen gibt.
     */
    public boolean hasElements() {
        return !elements.isEmpty();
    }

    public Hashtable getElements() {
        return elements;
    }

    public Iterator scopeIdentifiers() {
        return new Iterator() {
            private Iterator i = identifiers();
            private Definition sdCurrent, sdNext;

            public void remove() {
                throw new IllegalArgumentException();
            }

            public boolean hasNext() {
                return findNext(false);
            }

            public Object next() {
                if (!findNext(true)) {
                    throw new NoSuchElementException();
                }

                return sdCurrent;
            }

            private boolean findNext(boolean fGet) {
                if (sdNext != null) {
                    if (fGet) {
                        sdCurrent = sdNext;
                        sdNext = null;
                    }
                    return true;
                }

                while (i.hasNext()) {
                    Definition d = (Definition)i.next();
                    if (d instanceof ScopeDef) {
                        if (!fGet) {
                            sdNext = d;
                        } else {
                            sdCurrent = d;
                        }
                        return true;
                    }
                }
                return false;
            }

        };
    }

    public Iterator identifiers() {
         return new Iterator() {

            private Iterator i = elements.keySet().iterator();

            public void remove() {
                throw new IllegalArgumentException();
            }

            public boolean hasNext() {
                return i.hasNext();
            }

            public Object next() {
                return lookup((String)i.next());
            }

        };
    }

    /**
     * Sucht nach einer Definition in diesem Sichtbarkeitsbereich. Wenn es keine gibt,
     * wird rekursiv im Elternsichtbarkeitsbereich (sofern vorhanden) weitergesucht. Wird die
     * gesuchte Definition gefunden, wird sie zurckgeliefert, andernfalls wird null zurckgegeben.
     */
    Definition lookup(String name) {
        // Try to find the name in our scope
        Definition d = (Definition)elements.get(name);
        if (d == null && getParentScope() != null) {
            d = getParentScope().lookup(name);
        }
        return d;
    }

    public int getScopeNumber() {
        return scopeNumber;
    }

}
