package frontend.symbol;

import antlr.Token;

/**
 * Definiert ein Symbol, welches in die Symboltabelle eingetragen werden kann. Symbole haben einen Namen,
 * den Bezeichner, mit dem sie im Quelltext referenziert werden, und einen Typ (Konstante, Variable, ...).
 * Intern werden in der Symboltabelle allerdings sogenannte Definitions gespeichert. Die Klasse Symbol
 * ist ein Adapter vom Parser zur Symboltabelle. Der Parser generiert Symbole aus Tokens, welche von der
 * Symboltabelle in Definitions umgewandelt werden. Damit findet eine weitgehende Entkopplung von
 * Symboltabelle und Parser statt.
 */
public class Symbol {

    public static final int SYMBOL_CONST = 0;
    public static final int SYMBOL_VAR = 1;
    public static final int SYMBOL_PROC = 2;

    public static final String[] SYMBOLNAME_LC = new String[]{"constant", "variable", "procedure"};
    public static final String[] SYMBOLNAME_UC = new String[]{"Constant", "Variable", "Procedure"};

    /**
     * Name des Symbols
     */
    private String name;

    /**
     * Art des Symbols (Konstante, Variable, Prozedur)
     */
    private int type;

    /**
     * Zeile im Quellcode
     */
    private int line;

    /**
     * Spalte im Quellcode
     */
    private int column;


    /**
     * Konstruktor, wird benutzt bei Deklarationen, wenn der Typ bekannt ist.
     * @param Token der Token, aus dem das Symbol erzeugt wird.
     * @param type Typ des Symbols, kann SYMBOL_CONST, SYMBOL_VAR oder SYMBOL_PROC sein.
     */
    public Symbol(Token t, int type) {
        this(t.getText(), t.getLine(), t.getColumn(), type);
    }

    /**
     * Konstruktor, wird benutzt bei Vorkommen im Quelltext, wenn der Typ nicht bekannt ist.
     * @param t der Token, aus dem das Symbol erzeugt wird.
     */
    public Symbol(Token t) {
        this(t, -1);
    }

    public Symbol(String name, int line, int column, int type) {
        this.name = name;
        this.type = type;
        this.line = line;
        this.column = column;
    }


    /**
     * Liefert den Namen des Symbols zurck.
     */
    public String getName() {
        return name;
    }

    /**
     * Liefert den Typ des Symbols zurck.
     */
    public int getType() {
        return type;
    }

    /**
     * Liefert die Zeile des Symbols zurck.
     */
    public int getLine() {
        return line;
    }

    /**
     * Liefert die Spalte des Symbols zurck
     */
    public int getColumn() {
        return column;
    }

    public String toString() {
        return "(" + name + ", " + type + ")";
    }
}
