package frontend.symbol;

import java.util.*;

/**
 * Die Symboltabelle. Alle im Quellprogramm deklarierten Bezeichner werden hier eingetragen. Werden
 * nichtdeklarierende (also nutzende) Vorkommen der Bezeichner gefunden, so wird berprft, ob diese
 * bereits deklariert wurden. Das ist der Fall, wenn sie in der Symboltabelle enthalten sind. Werden die
 * Bezeichner nicht gefunden, so handelt es sich bei ihnen eventuell um Prozedurnamen, die spter definiert
 * werden. Es wird ein Platzhaltereintrag (UnknownDef) in der Liste unknowns gespeichert. Nach Durchlaufen
 * des Quelltextes wird resolveUnknown ausgefhrt und berprft, ob es zu den Platzhaltern mittlerweile
 * eine Deklaration gibt. Wenn nicht, wird eine NoDeclarationException geworfen.
 */
public class SymbolTable {

    /**
     * Stack mit Sichtbarkeitsbereichen. Da Sichtbarkeitsbereiche geschachtelt sind, bietet der Stack eine
     * einfache Mglichkeit, einen Sichtbarkeitsbereich hinzuzufgen oder zu entfernen. Der aktuelle
     * Sichtbarkeitsbereich liegt obenauf.
     */
    private ScopeStack scopeStack = new ScopeStack();

    private LinkedList unknowns = new LinkedList();

    private ScopeDef baseScope = new ProcedureDef("$main$", null, null);
    /**
     * Singletoninstanz der Symboltabelle
     */
    private static SymbolTable symbolTable = null;

    private static int scopeCounter = 0;

    /**
     * Privater Konstruktor der Symboltabelle, um Singleton zu erzwingen. Es wird der Basissichtbereich
     * definiert (in Form einer Prozedurdefinition, da das Gesamtprogramm in seinem Aufbau einer Prozedur
     * entspricht) und als aktueller Bereich auf den Sichtbarkeitsstack gelegt.
     */
    private SymbolTable() {
        scopeStack.push(baseScope);
    }

    /**
     * Gibt die Singletoninstanz der Symboltabelle zurck. Wann immer die Symboltabelle bentigt wird,
     * muss sie mit dieser Methode geholt werden.
     */
    public static SymbolTable getSymbolTable() {
        if (symbolTable == null) {
            symbolTable = new SymbolTable();
        }
        return symbolTable;
    }

    public void reset() {
        scopeCounter = 0;
        symbolTable = new SymbolTable();
    }

    public static int getScopeCounter() {
        return scopeCounter++;
    }

    public ScopeDef getBaseScope() {
        return baseScope;
    }



    /**
     * Schaut nach, ob ein bestimmter Bezeichner definiert ist, d.h., ob der schon in der Symboltabelle
     * eingetragen wurde. Dazu wird ein lookup auf dem aktuellen Sichtbarkeitsbereich durchgefhrt.
     */
    public Definition lookup(String s) {
        Definition def = scopeStack.lookup(s);
        return def;
    }

    public int[] lookupPos(String s) {
        return scopeStack.lookupPos(s);
    }

    /**
     * berprft, ob ein bestimmter Bezeichner bereits deklariert wurde. Ruft im Wesentlichen lookup auf,
     * wenn der Bezeichner nicht gefunden wurde, wird er als unbekannte Definition temporr gespeichert.
     * Nachdem der Syntaxbaum durchlaufen wurde, werden alle unbekannten Bezeichner noch einmal berprft.
     * Damit sind forward Deklarationen von Prozeduren mglich.
     */
    public void identifierUsage(Symbol s) {
        if (lookup(s.getName()) == null) {
            addUnknownDef(s);
        }
    }

    /**
     * Fgt ein Symbol zur Liste der noch nicht deklarierten Symbole hinzu. Position und Sichtbarkeitsbereich
     * werden festgehalten.
     */
    private void addUnknownDef(Symbol s) {
        Occurrence occ = new Occurrence(s.getLine(), s.getColumn());
        UnknownDef unknown = new UnknownDef(s.getName(), occ, getCurrentScope());
        unknowns.add(unknown);
    }

    /**
     * Wenn der Parser eine Prozedurdeklaration findet, wird sie mit dieser Methode in die Symboltabelle
     * eingetragen. Die Prozedur wird auerdem in den aktuellen Sichtbarkeitsbereich eingetragen. Da eine
     * Prozedur selbst einen neuen Sichtbarkeitsbereich definiert, wird sie oben auf den Stack gelegt und
     * ist damit der neue aktuelle Sichtbarkeitsbereich.
     */
    public void defineProcedure(Symbol s) {
        Occurrence occ = new Occurrence(s.getLine(), s.getColumn());
        ProcedureDef procedure = new ProcedureDef(s.getName(), occ, getCurrentScope());
        addToCurrentScope(procedure);
        pushScope(procedure);
    }

    /**
     * Wenn der Parser eine Variablendeklaration findet, wird sie mit dieser Methode in die Symboltabelle
     * eingetragen. Die Variable wird auerdem in den aktuellen Sichtbarkeitsbereich eingetragen.
     */
    public void defineVar(Symbol s) {
        Occurrence occ = new Occurrence(s.getLine(), s.getColumn());
        VariableDef variable = new VariableDef(s.getName(), occ, getCurrentScope());
        addToCurrentScope(variable);
    }

    /**
     * Wenn der Parser eine Konstantendeklaration findet, wird sie mit dieser Methode in die Symboltabelle
     * eingetragen. Die Konstante wird auerdem in den aktuellen Sichtbarkeitsbereich eingetragen.
     */
    public void defineConst(Symbol s, int value) {
        Occurrence occ = new Occurrence(s.getLine(), s.getColumn());
        ConstantDef constant = new ConstantDef(s.getName(), value, occ, getCurrentScope());
        addToCurrentScope(constant);
    }

    /**
     * Hilfsmethode, um eine Definition (Variable, Konstante, Prozedur) dem aktuellen Sichtbarkeitsbereich
     * hinzuzufgen. Dazu wird die Definition in die Liste aller sichtbaren Definitionen des Bereichs
     * aufgenommen. Anschlieend wird der Definition mitgeteilt, zu welchem Sichtbarkeitsbereich sie gehrt.
     */
    private void addToCurrentScope(Definition def) {
        getCurrentScope().add(def);
        def.setParentScope(getCurrentScope());
    }

    /**
     * Wenn smtliche Deklarationen gesammelt wurden, wird nachgeschaut, ob die undefinierten
     * Prozedurbezeichner mittlerweile innerhalb des Gltigkeitsbereichs deklariert wurden. Falls nicht,
     * wird eine NoDeclarationException geworfen.
     *
     * Achtung: Das funktioniert nur, wenn ausschlielich Prozeduren vor ihrer Deklaration benutzt werden
     * knnen (was zum Glck fr PL/0 gegeben ist). Gbe es noch andere Konstrukte, wo das mglich ist, msste
     * noch einmal berprft werden, ob der Typ des nun gefundenen Symbols der ist, der ursprnglich erwartet
     * wurde. (Was ursprnglich erwartet wurde, msste dann auch noch irgendwo gespeichert werden,
     * neue Klasse?)
     */
    public void resolveUnknown() {
        for (Iterator it = unknowns.iterator(); it.hasNext();) {
            UnknownDef ud = (UnknownDef)it.next();
            Definition looked = ud.getParentScope().lookup(ud.getName());
            Symbol s = new Symbol(ud.getName(), ud.getOccurrence().getLine(),
                    ud.getOccurrence().getColumn(), -1);
            if (looked == null) {
                throw new NoDeclarationException(s);
            }
        }
    }


    /**
     * Liefert den aktuellen Sichtbarkeitsbereich zurck.
     */
    private ScopeDef getCurrentScope() {
        if (scopeStack.empty())
            return null;
        return (ScopeDef)scopeStack.peek();
    }

    /**
     * Entfernt den aktuellen Sichtbarkeitsbereich vom Stack. Wird vom Parser aufgerufen, nachdem eine
     * Prozedurdefinition abgeschlossen und deren Sichtbarkeitsbereich zuende ist.
     */
    public void popScope() {
        scopeStack.pop();
    }


    /**
     *  Legt einen Sichtbarkeitsbereich auf den Stack. Wird von defineProcedure genutzt.
     */
    public void pushScope(ScopeDef scope) {
        scopeStack.push(scope);
    }


    public void checkForVariable(Symbol s) {
        Definition d = lookup(s.getName());
        if (d == null) {
            throw new NoDeclarationException(s);
        }
        if (!(d instanceof VariableDef)) {
            throw new NoVariableException(s.getName(), s.getLine(), s.getColumn(), d.getType());
        }
    }

    /**
     * berprft, ob ein Symbol fr einen Prozedurbezeichner steht. Dazu wird in der Symboltabelle
     * nachgeschlagen. Liefert die Symboltabelle keinen Eintrag, so wird die Deklaration unter Umstnden
     * erst noch folgen und das Symbol wird in die Liste mit den undeklarierten Symbolen (Liste unknowns)
     * eingefgt. Das Symbol wird am Ende des Parsevorgangs in der Methode resolveUnknowns erneut untersucht
     * werden. Wird beim Nachschlagen nicht null zurckgeliefert, steht fest, dass es eine Prozedur ist.
     *
     * Achtung: Das funktioniert nur, wenn ausschlielich Prozeduren vor ihrer Deklaration benutzt werden
     * knnen (was zum Glck fr PL/0 gegeben ist). Gbe es noch andere Konstrukte, wo das mglich ist, msste
     * in der Methode resolveUnknowns noch einmal berprft werden, ob der Typ des nun gefundenen Symbols
     * tatschlich der richtige ist.
     */
    public void checkForProcedure(Symbol s) {
        Definition d = lookup(s.getName());
        if (d == null) {
            addUnknownDef(s);
        } else {
            if (!(d instanceof ProcedureDef)) {
                throw new NoProcedureException(s.getName(), s.getLine(), s.getColumn(), d.getType());
            }
        }

    }



    public void checkForValue(Symbol s) {
        Definition d = lookup(s.getName());
        if (d == null) {
            throw new NoDeclarationException(s);
        }
        if (!(d instanceof VariableDef) && !(d instanceof ConstantDef)) {
            throw new NoValueException(s.getName(), s.getLine(), s.getColumn(), d.getType());
        }
    }


}